using UnityEngine;
using UnityEngine.UI;
using UnityEngine.Events;

using SimpleJSON;
using System.Collections.Generic;

using MacGruber;

namespace ChenRong
{
    public class CanvasButton : MVRScript
    {
        #region PluginInfo
        public string pluginAuthor = "ChenRong";
        public string pluginName = "CanvasButton";
        public string pluginVersion = "2.1";
        public string pluginDate = "2024-01-18";
        public string pluginDescription = @"Canvas Button Plugin";
        #endregion


        private JSONStorableString ui_name;
        private JSONStorableString new_name;

        private JSONStorableFloat spacing = null;
        private JSONStorableFloat[] size = new JSONStorableFloat[2];
        private JSONStorableFloat[] position = new JSONStorableFloat[3];

        private JSONStorableStringChooser ui_chooser;
        private List<object> ui_objects = new List<object>();

        private bool active = true;
        private static Canvas canvas = null;
        private JSONStorableBool is_panelopen;
        private JSONStorableBool is_horizontally;

        // private Color text_color = new Color(0, 0, 0);
        // private Color button_color = new Color(0.8392f, 0.8392f, 0.8392f);

        JSONStorableColor text_color = null;
        JSONStorableColor button_color = null;

        private JSONStorableFloat button_alpha = null;

        private List<string> names = new List<string>();
        private Dictionary<string, string> labels = new Dictionary<string, string>();
        private Dictionary<string, EventTrigger> events = new Dictionary<string, EventTrigger>();
        private Dictionary<string, UIDynamicButton> buttons =
            new Dictionary<string, UIDynamicButton>();
        private Dictionary<string, JSONStorableAction> handles =
            new Dictionary<string, JSONStorableAction>();

        public override void Init()
        {
            SimpleTriggerHandler.LoadAssets();

            generate_variable();
            generate_uiplugin();
        }

        public void Start()
        {
            this.ui_name.valNoCallback = "Action 1";
            this.new_name.valNoCallback = "NewAction";

            restore_buttons();
            restore_uiplugin();

            SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
        }

        private void Update()
        {
            foreach (var name in this.names)
                this.events[name].Update();
        }

        void OnDestroy()
        {
            clear_canvas();

            SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
        }

        private void set_active(bool active)
        {
            this.active = active;
            foreach (var name in this.names)
                this.buttons[name].gameObject.SetActive(active);
        }

        private string create_name(string name)
        {
            var str = System.Text.RegularExpressions.Regex.Replace(name, @"[^0-9]+", "");
            if (str.Length > 0)
            {
                var no = int.Parse(str);
                var newn = name.Substring(0, name.Length - str.Length) + (no + 1).ToString();
                return newn;
            }
            else
                return name + " 1";
        }

        private void generate_button()
        {
            var name = this.ui_name.val;
            var action = new JSONStorableAction(
                name,
                () =>
                {
                    this.events[name].Trigger();
                }
            );

            this.names.Add(name);
            this.labels.Add(name, name);
            this.events.Add(name, new EventTrigger(this, name));
            this.handles.Add(name, action);
            this.buttons.Add(
                name,
                add_button(
                    name,
                    name,
                    () =>
                    {
                        action.actionCallback();
                    },
                    this.names.Count - 1,
                    is_horizontally.val
                )
            );
            if (is_panelopen.val)
                this.events[name].OpenPanel();

            restore_uiplugin();
            sync_trigger_trace();
            RegisterAction(action);

            this.ui_name.valNoCallback = create_name(name);
        }

        private void rename_button()
        {
            var name = "";
            foreach (var n in this.names)
            {
                if (this.buttons[n].label == this.new_name.val)
                {
                    SuperController.LogError(
                        $"Name " + this.new_name.val + " is illegal, please rename it"
                    );
                    return;
                }

                // Fixed rename button twice or more
                if (this.buttons[n].label == this.ui_chooser.val)
                    name = n;
            }

            // var name = this.ui_chooser.val;
            var new_name = this.new_name.val;
            this.labels[name] = new_name;
            DeregisterAction(this.handles[name]);
            this.handles.Remove(name);

            var action = new JSONStorableAction(
                new_name,
                () =>
                {
                    this.events[name].Trigger();
                }
            );
            RegisterAction(action);
            this.handles.Add(name, action);

            restore_canvas();
            restore_uiplugin();
        }

        private void destroy_button()
        {
            var label = this.ui_chooser.val;
            foreach (var name in this.names)
            {
                if (this.buttons[name].label != label)
                    continue;
                delete_button(name);

                this.names.Remove(name);
                // this.events[name].Remove();
                this.events.Remove(name);
                this.buttons.Remove(name);

                restore_canvas();
                restore_uiplugin();
            }
        }

        private void destroy_buttons()
        {
            foreach (var name in this.names)
                delete_button(name);

            this.names.Clear();
            this.labels.Clear();
            this.events.Clear();
            this.buttons.Clear();

            this.ui_name.valNoCallback = "Action 1";
            this.new_name.valNoCallback = "NewAction";

            restore_canvas();
            restore_uiplugin();
        }

        private void restore_buttons()
        {
            foreach (var name in this.names)
            {
                this.handles.Add(
                    name,
                    new JSONStorableAction(
                        name,
                        () =>
                        {
                            this.events[name].Trigger();
                        }
                    )
                );
            }

            restore_canvas();
        }

        private void generate_canvas()
        {
            GameObject canvasObject = new GameObject();
            canvas = canvasObject.AddComponent<Canvas>();
            canvas.renderMode = RenderMode.WorldSpace;
            canvas.pixelPerfect = false;
            SuperController.singleton.AddCanvas(canvas);

            canvas.transform.SetParent(SuperController.singleton.mainHUD, false);

            CanvasScaler cs = canvasObject.AddComponent<CanvasScaler>();
            cs.scaleFactor = 80.0f;
            cs.dynamicPixelsPerUnit = 1f;

            GraphicRaycaster gr = canvasObject.AddComponent<GraphicRaycaster>(); // click enable

            float scale = 0.001f;
            canvas.transform.localScale = new Vector3(scale, scale, scale);
            canvas.transform.localPosition = new Vector3(0, 0, 0);
            canvas.transform.localEulerAngles = new Vector3(0, 180, 0); // look at camera
            canvas.transform.Translate(0, 0, 0);
        }

        private void clear_canvas()
        {
            if (canvas == null)
                return;

            if (SuperController.singleton != null)
                SuperController.singleton.RemoveCanvas(canvas);

            canvas.transform.SetParent(null, false);

            if (canvas.gameObject != null)
                GameObject.Destroy(canvas.gameObject);
        }

        private void restore_canvas()
        {
            clear_canvas();
            generate_canvas();

            this.buttons = new Dictionary<string, UIDynamicButton>();
            for (int i = 0; i < this.names.Count; i++)
            {
                var name = this.names[i];
                this.buttons.Add(
                    name,
                    add_button(
                        name,
                        this.labels[name],
                        () =>
                        {
                            this.handles[name].actionCallback();
                        },
                        i,
                        is_horizontally.val
                    )
                );
            }
        }

        public UIDynamicButton create_button(
            string name,
            UnityAction callback,
            float width = 100,
            float height = 80
        )
        {
            Transform button = GameObject.Instantiate<Transform>(
                this.manager.configurableButtonPrefab
            );
            ConfigureTransform(button, width, height);
            button.SetParent(canvas.transform, false); // ParentToCanvas

            UIDynamicButton uiButton = button.GetComponent<UIDynamicButton>();
            uiButton.label = name;
            uiButton.buttonText.fontSize = 20;
            uiButton.button.onClick.AddListener(callback);

            return uiButton;
        }

        private UIDynamicButton add_button(
            string name,
            string label,
            UnityAction callback,
            int index,
            bool direction
        )
        {
            float position_x = this.position[0].val * 2.6f;
            float position_y = -this.position[1].val * 1.4f;
            float position_z = -this.position[2].val;
            float column = 0;
            float row = 0;
            if (direction == true)
                column = -index * spacing.val * 3;
            else
                row = -index * spacing.val;

            UIDynamicButton button = create_button(name, callback, size[0].val, size[1].val);
            button.transform.Translate(
                -0.73f + position_x + column,
                1f + position_y + row,
                0.33f + position_z,
                Space.Self
            );

            button.label = label;

            Color c = HSVColorPicker.HSVToRGB(this.button_color.val);
            c.a = this.button_alpha.val;
            button.textColor = HSVColorPicker.HSVToRGB(this.text_color.val);
            button.buttonColor = c;

            return button;
        }

        private void delete_button(string name)
        {
            var action = this.handles[name];
            DeregisterAction(action);

            this.handles.Remove(name);
            this.events[name].Remove();

            UIDynamicButton button = this.buttons[name];
            Object.Destroy(button.gameObject);
        }

        private void sync_trigger_trace()
        {
            SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
            SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
        }

        private void generate_variable()
        {
            this.ui_name = new JSONStorableString("Name", "");
            this.new_name = new JSONStorableString("Name", "");

            this.position[0] = new JSONStorableFloat(
                "Position (x)",
                1f,
                (float v) =>
                {
                    restore_canvas();
                },
                0f,
                1f,
                false,
                true
            );
            this.position[1] = new JSONStorableFloat(
                "Position (y)",
                0f,
                (float v) =>
                {
                    restore_canvas();
                },
                0f,
                1f,
                false,
                true
            );
            this.position[2] = new JSONStorableFloat(
                "Position (z)",
                0f,
                (float v) =>
                {
                    restore_canvas();
                },
                -1f,
                1f,
                false,
                true
            );

            this.spacing = new JSONStorableFloat(
                "Spacing",
                0.05f,
                (float v) =>
                {
                    restore_canvas();
                },
                0f,
                0.2f,
                false,
                true
            );

            this.size[0] = new JSONStorableFloat(
                "Width",
                40,
                (float v) =>
                {
                    restore_canvas();
                },
                20,
                80,
                false,
                true
            );

            this.size[1] = new JSONStorableFloat(
                "Height",
                40,
                (float v) =>
                {
                    restore_canvas();
                },
                30,
                120,
                false,
                true
            );

            button_alpha = new JSONStorableFloat(
                "Alpha",
                1f,
                (float v) =>
                {
                    // this.button_color.a = v;
                    restore_canvas();
                },
                0f,
                1f,
                false,
                true
            );

            is_horizontally = new JSONStorableBool(
                "Horizontally Button",
                false,
                (bool v) =>
                {
                    restore_canvas();
                }
            );
            is_horizontally.storeType = JSONStorableParam.StoreType.Full;
            RegisterBool(is_horizontally);

            RegisterFloat(this.button_alpha);
            RegisterFloat(this.spacing);
            foreach (var v in this.position)
                RegisterFloat(v);
            foreach (var v in this.size)
                RegisterFloat(v);

            this.text_color = new JSONStorableColor("text_color", new HSVColor());
            this.button_color = new JSONStorableColor("button_color", new HSVColor());
            this.button_color.SetColor(new Color(0.8392f, 0.8392f, 0.8392f));

            RegisterColor(this.text_color);
            RegisterColor(this.button_color);

            this.ui_chooser = new JSONStorableStringChooser(
                "Selected State",
                new List<string>(),
                string.Empty,
                "Selected Button"
            );
            this.ui_chooser.displayChoices = new List<string>();

            RegisterAction(
                new JSONStorableAction(
                    "Toggle",
                    () =>
                    {
                        set_active(!active);
                    }
                )
            );
            RegisterAction(
                new JSONStorableAction(
                    "Show",
                    () =>
                    {
                        set_active(true);
                    }
                )
            );
            RegisterAction(
                new JSONStorableAction(
                    "Hide",
                    () =>
                    {
                        set_active(false);
                    }
                )
            );
        }

        private void generate_uiplugin()
        {
            Utils.OnInitUI(CreateUIElement);

            Utils.SetupInfoText(
                this,
                "<color=#000><size=35><b>CanvasButton</b></size></color>\n"
                    + "<color=#333>"
                    + "Add Button to 2D UI interface\n"
                    + "</color>",
                120.0f,
                true
            );

            Utils.SetupInfoOneLine(this, "<size=30><b>Button Config</b></size>", false);
            Utils.SetupTextInput(this, "Name", this.ui_name, false);

            Utils.SetupSpacer(this, 3, false);
            JSONStorableBool is_text_color = Utils.SetupToggle(
                this,
                "Button Text Color",
                false,
                false
            );

            JSONStorableColor color = Utils.SetupColor(
                this,
                "Button Color",
                // new Color(0.8392f, 0.8392f, 0.8392f),
                Color.HSVToRGB(
                    this.button_color.val.H,
                    this.button_color.val.S,
                    this.button_color.val.V
                ),
                false
            );
            color.setCallbackFunction += (float h, float s, float v) =>
            {
                if (is_text_color.val)
                {
                    this.text_color.SetVal(h, s, v);
                    foreach (var name in this.names)
                        this.buttons[name].textColor = HSVColorPicker.HSVToRGB(this.text_color.val);
                }
                else
                {
                    this.button_color.SetVal(h, s, v);
                    Color c = HSVColorPicker.HSVToRGB(this.button_color.val);
                    c.a = this.button_alpha.val;
                    foreach (var name in this.names)
                        this.buttons[name].buttonColor = c;
                }
            };

            Utils.SetupSpacer(this, 3, false);

            Utils.SetupButton(
                this,
                "Generate Button",
                () =>
                {
                    generate_button();
                },
                false
            );
            Utils.SetupTwinButton(
                this,
                "Remove Buttons",
                () =>
                {
                    destroy_buttons();
                },
                "Toggle Buttons",
                () =>
                {
                    set_active(!active);
                },
                false
            );

            Utils.SetupSpacer(this, 5, false);

            Utils.SetupInfoOneLine(this, "<size=30><b>Button Quick Trigger</b></size>", false);

            Utils.SetupInfoOneLine(this, "<size=30><b>Button Information</b></size>", true);
            CreateFilterablePopup(this.ui_chooser, true);

            Utils.SetupTextInput(this, "New Name", this.new_name, true);
            Utils.SetupTwinButton(
                this,
                "Update",
                () =>
                {
                    rename_button();
                },
                "Remove",
                () =>
                {
                    destroy_button();
                },
                true
            );

            Utils.SetupSpacer(this, 20, true);

            Utils.SetupInfoOneLine(this, "<size=30><b>Button Extra Config</b></size>", true);
            is_panelopen = Utils.SetupToggle(this, "Open Panel When Creating", true, true);
            CreateToggle(is_horizontally, true);

            CreateSlider(this.spacing, true);
            CreateSlider(this.button_alpha, true);
            foreach (var v in this.position)
                CreateSlider(v, true);
            foreach (var v in this.size)
                CreateSlider(v, true);
        }

        private void restore_uiplugin()
        {
            var choices = new List<string>();

            Utils.RemoveUIElements(this, ui_objects);

            foreach (var name in this.names)
            {
                ui_objects.Add(
                    Utils.SetupButton(
                        this,
                        "[ " + this.buttons[name].label + " ]   Trigger",
                        () =>
                        {
                            this.events[name].OpenPanel();
                        },
                        false
                    )
                );
                choices.Add(this.buttons[name].label);
            }

            this.ui_chooser.choices = null;
            this.ui_chooser.choices = choices;
            this.ui_chooser.displayChoices = choices;
            this.ui_chooser.val = choices[0];
        }

        private void OnAtomRename(string oldid, string newid)
        {
            foreach (var name in this.names)
                this.events[name].SyncAtomNames();
        }

        private void ConfigureTransform(Transform t, float width, float height)
        {
            t.transform.position = Vector3.zero;
            RectTransform rt = t.GetComponent<RectTransform>();
            rt.anchoredPosition = new Vector2(width / 2, height / 2);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
        }

        public override JSONClass GetJSON(
            bool includePhysical = true,
            bool includeAppearance = true,
            bool forceStore = false
        )
        {
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
            if (includePhysical || forceStore)
            {
                needsStore = true;

                jc["js_button_counts"] = this.names.Count.ToString();
                for (int i = 0; i < this.names.Count; i++)
                {
                    var name = this.names[i];
                    jc["js_button_name_" + i] = name;
                    jc["js_button_label_" + i] = this.labels[name];
                    jc[this.events[name].Name] = this.events[name].GetJSON(base.subScenePrefix);
                }
            }
            return jc;
        }

        public override void RestoreFromJSON(
            JSONClass jc,
            bool restorePhysical = true,
            bool restoreAppearance = true,
            JSONArray presetAtoms = null,
            bool setMissingToDefault = true
        )
        {
            base.RestoreFromJSON(
                jc,
                restorePhysical,
                restoreAppearance,
                presetAtoms,
                setMissingToDefault
            );
            if (jc.HasKey("js_button_counts"))
                for (int i = 0; i < int.Parse(jc["js_button_counts"]); i++)
                    if (jc.HasKey("js_button_name_" + i))
                    {
                        string name = jc["js_button_name_" + i];
                        this.names.Add(name);
                        this.labels.Add(name, jc["js_button_label_" + i]);
                        this.events.Add(name, new EventTrigger(this, name));
                    }

            base.RestoreFromJSON(
                jc,
                restorePhysical,
                restoreAppearance,
                presetAtoms,
                setMissingToDefault
            );
        }

        public override void LateRestoreFromJSON(
            JSONClass jc,
            bool restorePhysical = true,
            bool restoreAppearance = true,
            bool setMissingToDefault = true
        )
        {
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
            if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
            {
                // SuperController.LogMessage(">> 1 RestoreFromJSON " + jc.ToString());

                if (jc.HasKey("js_button_counts"))
                {
                    for (int i = 0; i < int.Parse(jc["js_button_counts"]); i++)
                    {
                        if (jc.HasKey("js_button_name_" + i))
                        {
                            string name = jc["js_button_name_" + i];
                            this.events[name].RestoreFromJSON(
                                jc,
                                base.subScenePrefix,
                                base.mergeRestore,
                                setMissingToDefault
                            );
                        }
                    }
                }
            }
        }
    }
}
